import numpy as np

from nacl.models import constraints

def test_salt2like_constraints(nsn=5, seed=0):
    """
    """
    pass


def test_salt2like_constraint_derivatives(nsn=5, seed=0):
    """
    """
    pass

def test_salt2like_constraints_deactivate(nsn=5, seed=0):
    """
    """
    pass
